const math = @import("./math.zig");
const meta = @import("./meta.zig");
const stack = @import("./stack.zig");
const testing = @import("./testing.zig");

///
/// Allocation options for an [Allocator].
///
pub const Allocation = struct {
    existing: ?[*]u8,
    alignment: u29,
    size: usize,
};

///
/// Closure for dynamic memory allocation through the referenced allocator state machine capture.
///
pub const Allocator = meta.Function(Allocation, ?[*]u8);

///
/// [MakeError.OutOfMemory] if the requested amount of memory could not be allocated.
///
pub const MakeError = error {
    OutOfMemory,
};

///
/// Closure that captures a reference to readable resources like block devices, memory buffers,
/// network sockets, and more.
///
pub const Reader = meta.Function([]u8, usize);

///
/// Returns a state machine for lazily computing all `Element` components of a given source input
/// that match a delimiting pattern.
///
pub fn Spliterator(comptime Element: type) type {
    return struct {
        source: []const Element,
        delimiter: []const Element,

        const Self = @This();

        ///
        /// Returns `true` if there is more data to be processed, otherwise `false`.
        ///
        pub fn hasNext(self: Self) bool {
            return (self.source.len != 0);
        }

        ///
        /// Iterates on `self` and returns the next view of [Spliterator.source] that matches
        /// [Spliterator.delimiter], or `null` if there is no more data to be processed.
        ///
        pub fn next(self: *Self) ?[]const Element {
            if (!self.hasNext()) return null;

            if (self.delimiter.len == 0) {
                defer self.source = self.source[self.source.len .. self.source.len];

                return self.source[0 .. self.source.len];
            }

            if (findFirstOf(Element, self.source, self.delimiter, struct {
                fn testEquality(this: Element, that: Element) bool {
                    return this == that;
                }
            }.testEquality)) |head| {
                defer self.source = self.source[(head + self.delimiter.len) .. self.source.len];

                return self.source[0 .. head];
            }

            defer self.source = self.source[self.source.len .. self.source.len];

            return self.source;
        }
    };
}

test "Spliterator(u8)" {
    // Empty source.
    {
        var spliterator = Spliterator(u8){
            .source = "",
            .delimiter = " ",
        };

        try testing.expect(!spliterator.hasNext());
    }

    // Empty delimiter.
    {
        var spliterator = Spliterator(u8){
            .source = "aaa",
            .delimiter = "",
        };

        try testing.expect(spliterator.hasNext());
        try testing.expect(equals(u8, spliterator.next().?, "aaa"));
        try testing.expect(!spliterator.hasNext());
    }

    // Single-character delimiter.
    {
        var spliterator = Spliterator(u8){
            .source = "single.character.separated.hello.world",
            .delimiter = ".",
        };

        const components = [_][]const u8{"single",
            "character", "separated", "hello", "world"};

        var index = @as(usize, 0);
        const components_tail = components.len - 1;

        while (spliterator.next()) |split| : (index += 1) {
            try testing.expect(spliterator.hasNext() == (index < components_tail));
            try testing.expect(equals(u8, split, components[index]));
        }

        try testing.expect(!spliterator.hasNext());
    }

    // Multi-character delimiter.
    {
        var spliterator = Spliterator(u8){
            .source = "finding a needle in a needle stack",
            .delimiter = "needle",
        };

        const components = [_][]const u8{"finding a ", " in a ", " stack"};
        var index = @as(usize, 0);
        const components_tail = components.len - 1;

        while (spliterator.next()) |split| : (index += 1) {
            try testing.expect(spliterator.hasNext() == (index < components_tail));
            try testing.expect(equals(u8, split, components[index]));
        }

        try testing.expect(!spliterator.hasNext());
    }
}

///
/// Closure that captures a reference to writable resources like block devices, memory buffers,
/// network sockets, and more.
///
pub const Writer = meta.Function([]const u8, usize);

///
/// Returns a sliced reference of the raw bytes in `pointer`.
///
pub fn bytesOf(pointer: anytype) switch (@typeInfo(@TypeOf(pointer))) {
    .Pointer => |info| if (info.is_const) []const u8 else []u8,
    else => @compileError("`pointer` must be a pointer type"),
} {
    const Pointer = @TypeOf(pointer);
    const pointer_info = @typeInfo(Pointer).Pointer;

    switch (pointer_info.size) {
        .Many => @compileError("`pointer` cannot be an unbound pointer type"),
        .C => @compileError("`pointer` cannot be a C-style pointer"),

        .One => return @ptrCast(if (pointer_info.is_const) [*]const u8
                else [*]u8, pointer)[0 .. @sizeOf(Pointer)],

        .Slice => return @ptrCast(if (pointer_info.is_const) [*]const u8 else
            [*]u8, pointer.ptr)[0 .. (@sizeOf(Pointer) * pointer.len)],
    }
}

test "bytesOf" {
    var foo: u32 = 10;

    try testing.expect(bytesOf(&foo)[0] == 0x0a);
}

///
/// Compares `this` to `that`, returning the difference between the first byte deviation in the two
/// sequences, otherwise `0` if they are identical.
///
pub fn compareBytes(this: []const u8, that: []const u8) isize {
    const range = math.min(usize, this.len, that.len);
    var index: usize = 0;

    while (index < range) : (index += 1) {
        const difference = (this[index] - that[index]);

        if (difference != 0) return difference;
    }

    return (@intCast(isize, this.len) - @intCast(isize, that.len));
}

test "compareBytes" {
    try testing.expect(compareBytes(&.{69, 42, 0}, &.{69, 42, 0}) == 0);
    try testing.expect(compareBytes(&.{69, 42, 11}, &.{69, 42}) == 1);
    try testing.expect(compareBytes(&.{69, 42}, &.{69, 42, 11}) == -1);
}

///
/// Copies the contents of `source` into `target`
///
pub fn copy(comptime Element: type, target: []Element, source: []const Element) void {
    for (source) |element, index| target[index] = element;
}

test "copy" {
    var buffer = [_]u32{0} ** 20;
    const data = [_]u32{3, 20, 8000};

    copy(u32, &buffer, &data);

    for (data) |datum, index| try testing.expect(buffer[index] == datum);
}

///
/// Returns `true` if `this` is the same length and contains the same data as `that`, otherwise
/// `false`.
///
pub fn equals(comptime Element: type, this: []const Element, that: []const Element) bool {
    if (this.len != that.len) return false;

    var index: usize = 0;

    while (index < this.len) : (index += 1) if (this[index] != that[index]) return false;

    return true;
}

test "equals" {
    const bytes_sequence = &.{69, 42, 0};

    try testing.expect(equals(u8, bytes_sequence, bytes_sequence));
    try testing.expect(!equals(u8, bytes_sequence, &.{69, 42}));
}

///
/// Fills the contents of `target` with `source`.
///
pub fn fill(comptime Element: type, target: []Element, source: Element) void {
    for (target) |_, index| target[index] = source;
}

test "fill" {
    var buffer = [_]u32{0} ** 8;

    fill(u32, &buffer, 1);

    for (buffer) |element| try testing.expect(element == 1);
}

///
/// Linearly searches for the first instance of an `Element` equal to `needle` in `haystack`,
/// returning its index or `null` if nothing was found.
///
/// **Note** that this operation has `O(n)` time complexity.
///
pub fn findFirst(comptime Element: type, haystack: []const Element,
    needle: Element, comptime testEquality: fn (Element, Element) bool) ?usize {

    for (haystack) |element, index| if (testEquality(element, needle)) return index;

    return null;
}

test "findFirst" {
    const haystack = &.{"", "", "foo"};

    const testEquality = struct {
        fn testEquality(this: []const u8, that: []const u8) bool {
            return equals(u8, this, that);
        }
    }.testEquality;

    try testing.expect(findFirst([]const u8, haystack, "foo", testEquality).? == 2);
    try testing.expect(findFirst([]const u8, haystack, "bar", testEquality) == null);
}

///
/// Searches for the first instance of an `Element` sequence equal to the contents of `needle` in
/// `haystack`, returning the starting index or `null` if nothing was found.
///
/// **Note** that this operation has `O(nm)` time complexity.
///
pub fn findFirstOf(comptime Element: type, haystack: []const Element,
    needle: []const Element, comptime testEquality: fn (Element, Element) bool) ?usize {

    var head: usize = 0;
    const tail = (haystack.len - needle.len);

    walk_haystack: while (head <= tail) : (head += 1) {
        for (needle) |element, index|
            if (!testEquality(haystack[head + index], element)) continue: walk_haystack;

        return head;
    }

    return null;
}

test "findFirstOf" {
    const haystack = &.{"foo", "bar", "baz"};

    const testEquality = struct {
        fn testEquality(this: []const u8, that: []const u8) bool {
            return equals(u8, this, that);
        }
    }.testEquality;

    try testing.expect(findFirstOf([]const u8, haystack, &.{"bar", "baz"}, testEquality).? == 1);
    try testing.expect(findFirstOf([]const u8, haystack, &.{"baz", "bar"}, testEquality) == null);
}

///
/// Frees `allocated_memory` using `allocator`.
///
/// *Note* that only memory known to be freeable by `allocator` should be passed via
/// `allocated_memory`. Anything else will result is considered unreachable logic.
///
pub fn free(allocator: Allocator, allocated_memory: anytype) void {
    if (allocator.call(.{
        .existing = @ptrCast([*]u8, switch (@typeInfo(@TypeOf(allocated_memory))) {
            .Pointer => |info| switch (info.size) {
                .One, .Many, .C => allocated_memory,
                .Slice => allocated_memory.ptr,
            },

            else => @compileError("`allocated_memory` must be a pointer"),
        }),

        .size = 0,
        .alignment = 0,
    }) != null) unreachable;
}

test "free" {
    var buffer = [_]u8{0} ** 4096;
    var memory = stack.Fixed(u8){.buffer = &buffer};
    const fixed_allocator = stack.fixedAllocator(&memory);
    const block_size = 8;
    const allocated_block = (try makeMany(u8, fixed_allocator, block_size))[0 .. block_size];

    defer free(fixed_allocator, allocated_block);
}

///
/// Returns a deterministic hash code compiled from each byte in `bytes`.
///
/// **Note** that this operation has `O(n)` time complexity.
///
pub fn hashBytes(bytes: []const u8) usize {
    var hash = @as(usize, 5381);

    for (bytes) |byte| hash = ((hash << 5) + hash) + byte;

    return hash;
}

test "hashBytes" {
    const bytes_sequence = &.{69, 42, 0};

    try testing.expect(hashBytes(bytes_sequence) == hashBytes(bytes_sequence));
    try testing.expect(hashBytes(bytes_sequence) != hashBytes(&.{69, 42}));
}

///
/// Attempts to allocate a buffer of `size` `Element`s using `allocator`, returning it or a
/// [MakeError] if it failed.
///
pub fn makeMany(comptime Element: type, allocator: Allocator, size: usize) MakeError![*]Element {
    const alignment = @alignOf(Element);

    return @ptrCast([*]Element, @alignCast(alignment, allocator.call(.{
        .existing = null,
        .size = @sizeOf(Element) * size,
        .alignment = alignment,
    }) orelse return error.OutOfMemory));
}

test "makeMany" {
    var buffer = [_]u8{0} ** 4096;
    var memory = stack.Fixed(u8){.buffer = &buffer};
    const block_size = 8;

    // Don't care about the actual allocation - just assertions about it.
    _ = (try makeMany(u8, stack.fixedAllocator(&memory), block_size))[0 .. block_size];
}

///
/// Attempts to allocate a buffer of `1` `Element` using `allocator`, returning it or a [MakeError]
/// if it failed.
///
pub fn makeOne(comptime Element: type, allocator: Allocator) MakeError!*Element {
    const alignment = @alignOf(Element);

    return @ptrCast(*Element, @alignCast(alignment, allocator.call(.{
        .existing = null,
        .size = @sizeOf(Element),
        .alignment = alignment,
    }) orelse return error.OutOfMemory));
}

test "makeOne" {
    var buffer = [_]u8{0} ** 4096;
    var memory = stack.Fixed(u8){.buffer = &buffer};

    // Don't care about the actual allocation - just assertions about it.
    _ = try makeOne(u8, stack.fixedAllocator(&memory));
}

///
/// Swaps the `Data` in `this` with `that`.
///
pub fn swap(comptime Data: type, this: *Data, that: *Data) void {
    const temp = this.*;
    this.* = that.*;
    that.* = temp;
}

test "swap" {
    var a: u64 = 0;
    var b: u64 = 1;

    swap(u64, &a, &b);

    try testing.expect(a == 1);
    try testing.expect(b == 0);
}

///
/// Thread-safe and lock-free [Writer] that silently consumes all given data without failure and
/// throws it away.
///
/// This is commonly used for testing or redirected otherwise unwanted output data that has to be
/// sent somewhere for whatever reason.
///
pub const null_writer = Writer.from(struct {
    fn write(buffer: []const u8) usize {
        return buffer.len;
    }
}.write);

test "null_writer" {
    const sequence = "foo";

    try testing.expect(null_writer.call(sequence) == sequence.len);
}
