const io = @import("./io.zig");
const testing = @import("./testing.zig");

///
/// Returns a fixed-size stack type of `Element`s.
///
pub fn Fixed(comptime Element: type) type {
    return struct {
        filled: usize = 0,
        buffer: []Element,

        ///
        /// Stack type.
        ///
        const Self = @This();

        ///
        /// Resets the number of filled items to `0`, otherwise leaving the actual memory contents
        /// of the buffer untouched until it is later overwritten by following operations on it.
        ///
        pub fn clear(self: *Self) void {
            self.filled = 0;
        }

        ///
        /// Returns `true` if `self` has filled its buffer to maximum capacity, otherwise `false`.
        ///
        pub fn isFull(self: Self) bool {
            return (self.filled == self.buffer.len);
        }

        ///
        /// If `self` is filled with at least `1` value, it is decremented by `1`, otherwise leaving
        /// the actual memory contents of the buffer untouched until it is later overwritten by
        /// following operations on it.
        ///
        /// The value of the element removed from the list is returned if something existed to be
        /// popped, otherwise `null` if it contained no elements.
        ///
        pub fn pop(self: *Self) ?Element {
            if (self.filled == 0) return null;

            self.filled -= 1;

            return self.buffer[self.filled];
        }

        ///
        /// Attempts to push `element` into `self`, returning a [FixedPushError] if it failed.
        ///
        pub fn push(self: *Self, element: Element) FixedPushError!void {
            if (self.isFull()) return error.BufferOverflow;

            self.buffer[self.filled] = element;
            self.filled += 1;
        }

        ///
        /// Attempts to push all of `elements` into `self`, returning a [FixedPushError] if it
        /// failed.
        ///
        pub fn pushAll(self: *Self, elements: []const Element) FixedPushError!void {
            const filled = (self.filled + elements.len);

            if (filled > self.buffer.len) return error.BufferOverflow;

            io.copy(Element, self.buffer[self.filled ..], elements);

            self.filled = filled;
        }

        ///
        /// Attempts to push `count` instances of `element` into `self`, returning a
        /// [FixedPushError] if it failed.
        ///
        pub fn pushMany(self: *Self, element: Element, count: usize) FixedPushError!void {
            const filled = (self.filled + count);

            if (filled > self.buffer.len) return error.BufferOverflow;

            io.fill(Element, self.buffer[self.filled ..], element);

            self.filled = filled;
        }
    };
}

test "Fixed([]const u8)" {
    const default_value = "";
    var buffer = [_][]const u8{default_value} ** 4;
    var shopping_list = Fixed([]const u8){.buffer = &buffer};

    // Pop empty stack.
    {
        try testing.expect(shopping_list.pop() == null);
        try testing.expect(shopping_list.filled == 0);
        try testing.expect(shopping_list.buffer.ptr == &buffer);
        try testing.expect(shopping_list.buffer.len == buffer.len);

        for (shopping_list.buffer) |item|
            try testing.expect(io.equals(u8, item, default_value));
    }

    // Push single element.
    {
        try shopping_list.push("milk");
        try testing.expect(shopping_list.filled == 1);
        try testing.expect(shopping_list.buffer.ptr == &buffer);
        try testing.expect(shopping_list.buffer.len == buffer.len);
        try testing.expect(io.equals(u8, shopping_list.buffer[0], "milk"));

        for (shopping_list.buffer[1 ..]) |item|
            try testing.expect(io.equals(u8, item, default_value));

        // TODO: Test stack overflow.
    }

    // Pop single element.
    {
        try testing.expect(io.equals(u8, shopping_list.pop().?, "milk"));
        try testing.expect(shopping_list.filled == 0);
        try testing.expect(shopping_list.buffer.ptr == &buffer);
        try testing.expect(shopping_list.buffer.len == buffer.len);
        try testing.expect(io.equals(u8, shopping_list.buffer[0], "milk"));

        for (shopping_list.buffer[1 ..]) |item|
            try testing.expect(io.equals(u8, item, default_value));
    }

    // TODO: Multiple elements.
    // TODO: Clear elements.
}

///
/// Potential errors that may occur while trying to push one or more elements into a [Fixed] stack.
///
pub const FixedPushError = error {
    BufferOverflow,
};

///
/// Creates and returns a [io.Allocator] value wrapping `fixed_stack`.
///
/// The returned [io.Allocator] uses `fixed_stack` and its backing memory buffer as a fixed-length
/// memory pool to linearly allocate memory from.
///
pub fn fixedAllocator(fixed_stack: *Fixed(u8)) io.Allocator {
    return io.Allocator.fromClosure(fixed_stack, struct {
        fn alloc(stack: *Fixed(u8), allocation: io.Allocation) ?[*]u8 {
            if (allocation.existing) |buffer| if (allocation.size == 0) {
                // Deallocate the memory.
                const buffer_address = @ptrToInt(buffer);
                const stack_address = @ptrToInt(stack.buffer.ptr);

                // Check the buffer is within the address space of the stack buffer. If not, it
                // should just be returned to let the caller know it cannot be freed.
                if (buffer_address < stack_address or buffer_address >=
                    (stack_address + stack.filled)) return buffer;

                // TODO: Investigate ways of actually freeing if it is the last allocation.
                return null;
            } else {
                // TODO: Investigate ways of in-place relocating if it is the last allocation.
            };

            // Reallocate / allocate the memory.
            // TODO: Remove stdlib dependency.
            const adjusted_offset = @import("std").mem.alignPointerOffset(stack.buffer.ptr +
                stack.filled, allocation.alignment) orelse return null;

            const head = stack.filled + adjusted_offset;
            const tail = head + allocation.size;

            stack.pushMany(0, tail) catch return null;

            return stack.buffer[head .. tail].ptr;

        }
    }.alloc);
}

test "fixedAllocator" {
    var buffer = [_]u8{0} ** 32;
    var stack = Fixed(u8){.buffer = &buffer};
    const allocator = fixedAllocator(&stack);

    // Allocation
    var block_memory = allocator.call(.{
        .existing = null,
        .alignment = @alignOf(u64),
        .size = @sizeOf(u64),
    });

    try testing.expect(block_memory != null);

    const buffer_address_head = @ptrToInt(&buffer);
    const buffer_address_tail = @ptrToInt(&buffer) + buffer.len;

    {
        const block_memory_address = @ptrToInt(block_memory);

        try testing.expect(block_memory_address >= buffer_address_head and
            block_memory_address < buffer_address_tail);
    }

    // Reallocation.
    block_memory = allocator.call(.{
        .existing = block_memory,
        .alignment = @alignOf(u64),
        .size = @sizeOf(u64),
    });

    try testing.expect(block_memory != null);

    {
        const block_memory_address = @ptrToInt(block_memory);

        try testing.expect(block_memory_address >= buffer_address_head and
            block_memory_address < buffer_address_tail);
    }

    // Deallocation.
    try testing.expect(allocator.call(.{
        .existing = block_memory,
        .alignment = 0,
        .size = 0,
    }) == null);
}

///
/// Returns an [io.Writer] wrapping `fixed_stack`.
///
/// Writing to the returned [io.Writer] will push values to the underlying [Fixed] stack instance
/// referenced by `fixed_stack` until it is full.
///
pub fn fixedWriter(fixed_stack: *Fixed(u8)) io.Writer {
    return io.Writer.fromClosure(fixed_stack, struct {
        fn write(stack: *Fixed(u8), buffer: []const u8) usize {
            stack.pushAll(buffer) catch |err| switch (err) {
                error.BufferOverflow => return 0,
            };

            return buffer.len;
        }
    }.write);
}

test "fixedWriter" {
    var buffer = [_]u8{0} ** 4;
    var sequence_stack = Fixed(u8){.buffer = &buffer};
    const sequence_data = [_]u8{8, 16, 32, 64};

    try testing.expect(fixedWriter(&sequence_stack).call(&sequence_data) == sequence_data.len);
    try testing.expect(io.equals(u8, sequence_stack.buffer, &sequence_data));
}
