const core = @import("core");
const sys = @import("./sys.zig");

///
/// Metadata of an Oar archive entry.
///
const Block = extern struct {
    signature: [signature_magic.len]u8 = signature_magic,
    path: sys.Path = sys.Path.empty,
    data_size: u64 = 0,
    data_head: u64 = 0,
    padding: [232]u8 = [_]u8{0} ** 232,

    comptime {
        const entry_size = @sizeOf(@This());

        if (entry_size != 512) @compileError("EntryBlock is not 512 bytes");
    }
};

///
/// Reference to a file entry in an Oar archive, denoting the starting offset from the top of head
/// of the file and its size.
///
pub const Entry = struct {
    head: u64,
    size: u64,

    ///
    /// [FindError.EntryNotFound] occurs when no entry matched the parameters of the find operation.
    ///
    /// [FindError.ArchiveUnsupported] occurs if the file provided to the find operation is not a
    /// valid archive file.
    ///
    pub const FindError = error {
        EntryNotFound,
        ArchiveUnsupported,
    };

    ///
    /// Attempts to perform a binary search on the entry blocks defined in `archive_file` for one
    /// matching `entry_path`, returning an [Entry] referencing its data or a [FindError] if it
    /// failed.
    ///
    /// **Note** that this operation has `O(log n)` time complexity.
    ///
    pub fn find(archive_file: *sys.ReadableFile, entry_path: sys.Path) FindError!Entry {
        var header = Header{};
        const header_size = @sizeOf(Header);
        const io = core.io;

        if (((archive_file.read(0, io.bytesOf(&header)) catch
            return error.ArchiveUnsupported) != header_size) or
                (!io.equals(u8, &header.signature, &signature_magic)) or
                (header.revision != revision_magic) or
                (header.entry_head <= header_size)) return error.ArchiveUnsupported;

        // Read file table.
        var head: usize = 0;
        var tail: usize = (header.entry_count - 1);
        const block_size = @sizeOf(Block);

        while (head <= tail) {
            var block = Block{};
            const midpoint = (head + (tail - head) / 2);

            if ((archive_file.read(header.entry_head + (block_size * midpoint), io.bytesOf(&block))
                catch return error.ArchiveUnsupported) != block_size) return error.EntryNotFound;

            const comparison = entry_path.compare(block.path);

            if (comparison == 0) return Entry{
                .head = block.data_head,
                .size = block.data_size,
            };

            if (comparison > 0) {
                head = (midpoint + 1);
            } else {
                tail = (midpoint - 1);
            }
        }

        return error.EntryNotFound;
    }

    ///
    /// Reads the data from `entry` in `archive_file` from the byte at the entry-relative `offset`
    /// into `buffer` until either the end of the entry data, end of archive file, or end of buffer
    /// is reached.
    ///
    /// The number of bytes read is returned or [sys.FileError] if it failed.
    ///
    pub fn read(entry: Entry, archive_file: *sys.ReadableFile,
        offset: u64, buffer: []u8) sys.FileError!usize {

        return archive_file.read(entry.head + offset,
            buffer[0 .. core.math.min(usize, buffer.len, entry.size)]);
    }
};

///
/// Header data that every Oar archive file starts with at byte offset `0`.
///
const Header = extern struct {
    signature: [signature_magic.len]u8 = signature_magic,
    revision: u8 = revision_magic,
    entry_count: u32 = 0,
    entry_head: u64 = 0,
    padding: [496]u8 = [_]u8{0} ** 496,

    comptime {
        const size = @sizeOf(@This());

        if (size != 512) @compileError("Header is not 512 bytes");
    }
};

///
/// The magic revision number that this Oar software implementation understands.
///
const revision_magic = 0;

///
/// Magic identifier used to validate [Header] and [Block] data.
///
const signature_magic = [3]u8{'o', 'a', 'r'};
