const ext = @cImport({
    @cInclude("SDL2/SDL.h");
});

const core = @import("core");
const oar = @import("./oar.zig");
const std = @import("std");

///
/// Thread-safe platform abstraction over multiplexing system I/O processing and event handling.
///
pub const App = struct {
    message_chain: ?*Message = null,
    message_semaphore: *ext.SDL_sem,
    message_mutex: *ext.SDL_mutex,
    data: FileSystem,
    user: FileSystem,

    ///
    /// Enqueues `message` to the message chain in `app`.
    ///
    fn enqueue(app: *App, message: *Message) void {
        {
            // TODO: Error check these.
            _ = ext.SDL_LockMutex(app.message_mutex);

            defer _ = ext.SDL_UnlockMutex(app.message_mutex);

            if (app.message_chain) |message_chain| {
                message_chain.next = message;
            } else {
                app.message_chain = message;
            }
        }

        // TODO: Error check this.
        _ = ext.SDL_SemPost(app.message_semaphore);
    }

    ///
    /// Asynchronously executes `procedure` with `arguments` as an anonymous struct of its arguments
    /// and `app` as its execution context.
    ///
    /// Once the execution frame resumes, the value returned by executing `procedure` is returned.
    ///
    pub fn schedule(app: *App, procedure: anytype,
        arguments: anytype) core.meta.FnReturn(@TypeOf(procedure)) {

        const Task = struct {
            procedure: @TypeOf(procedure),
            arguments: *@TypeOf(arguments),
            result: core.meta.FnReturn(@TypeOf(procedure)),

            const Task = @This();

            fn process(userdata: *anyopaque) void {
                const task = @ptrCast(*Task, @alignCast(@alignOf(Task), userdata));

                task.result = @call(.{}, task.procedure, task.arguments.*);
            }
        };

        var task = Task{
            .procedure = procedure,
            .arguments = &arguments,
        };

        var message = Message{
            .kind = .{.task = .{
                .data = &task,
                .action = Task.process,
                .frame = @frame(),
            }},
        };

        suspend app.enqueue(&message);
    }

    ///
    /// Asynchronously logs `info` with `logger` as the logging method and `app` as the execution
    /// context.
    ///
    pub fn log(app: *App, logger: Logger, info: []const u8) void {
        var message = Message{
            .kind = .{.log = .{
                .logger = logger,
                .info = info,
            }},
        };

        app.enqueue(&message);
    }
};

///
///
///
pub const ReadableFile = opaque {
    ///
    ///
    ///
    pub fn close(readable_file: *ReadableFile) void {
        if (ext.SDL_RWclose(readable_file.rwOpsCast()) != 0)
            return ext.SDL_LogWarn(ext.SDL_LOG_CATEGORY_APPLICATION,
                "Attempt to close an invalid file reference");
    }

    ///
    ///
    ///
    pub fn read(readable_file: *ReadableFile, offset: u64, buffer: []u8) FileError!u64 {
        const rw_ops = readable_file.rwOpsCast();

        {
            ext.SDL_ClearError();

            const math = core.math;
            const min = math.min;
            const maxIntValue = math.maxIntValue;
            var sought = min(u64, offset, maxIntValue(i64));

            if (ext.SDL_RWseek(rw_ops, @intCast(i64, sought), ext.RW_SEEK_SET) < 0)
                return error.FileInaccessible;

            var to_seek = offset - sought;

            while (to_seek != 0) {
                sought = min(u64, to_seek, maxIntValue(i64));

                ext.SDL_ClearError();

                if (ext.SDL_RWseek(rw_ops, @intCast(i64, sought), ext.RW_SEEK_CUR) < 0)
                    return error.FileInaccessible;

                to_seek -= sought;
            }
        }

        ext.SDL_ClearError();

        const buffer_read = ext.SDL_RWread(rw_ops, buffer.ptr, @sizeOf(u8), buffer.len);

        if ((buffer_read == 0) and (ext.SDL_GetError() != null))
            return error.FileInaccessible;

        return buffer_read;
    }

    ///
    ///
    ///
    pub fn rwOpsCast(readable_file: *ReadableFile) *ext.SDL_RWops {
        return @ptrCast(*ext.SDL_RWops, @alignCast(@alignOf(ext.SDL_RWops), readable_file));
    }

    ///
    ///
    ///
    pub fn size(readable_file: *ReadableFile) FileError!u64 {
        ext.SDL_ClearError();

        const byte_size = ext.SDL_RWsize(readable_file.rwOpsCast());

        if (byte_size < 0) return error.FileInaccessible;

        return @intCast(u64, byte_size);
    }
};

///
/// [Error.FileInaccessible] is a generic catch-all for a [FileAccess] reference no longer pointing
/// to a file or the file becomming invalid for whatever reason.
///
pub const FileError = error {
    FileInaccessible,
};

///
/// Platform-agnostic mechanism for working with an abstraction of the underlying file-system(s)
/// available to the application in a sandboxed environment.
///
pub const FileSystem = union(enum) {
    native: []const u8,
    archive_file: *ReadableFile,

    ///
    /// With files typically being backed by a block device, they can produce a variety of errors -
    /// from physical to virtual errors - these are all encapsulated by the API as general
    /// [OpenError.FileNotFound] errors.
    ///
    /// When a given [FileSystem] does not support a specified [OpenMode],
    /// [OpenError.ModeUnsupported] is used to inform the consuming code that another [OpenMode]
    /// should be tried or, if no mode other is suitable, that the resource is effectively
    /// unavailable.
    ///
    /// If the number of known [FileAccess] handles has been exhausted, [OpenError.OutOfFiles] is
    /// used to communicate this.
    ///
    pub const OpenError = error {
        FileNotFound,
        ModeUnsupported,
        OutOfFiles,
    };

    ///
    /// Attempts to open the file identified by `path` with `mode` as the mode for opening the file.
    ///
    /// Returns a [ReadableFile] reference that provides access to the file referenced by `path`or a
    /// [OpenError] if it failed.
    ///
    pub fn openRead(file_system: *const FileSystem, path: Path) OpenError!*ReadableFile {
        switch (file_system.*) {
            .archive_file => |archive_file| {
                const entry = oar.Entry.find(archive_file, path) catch return error.FileNotFound;

                _ = entry;
                // TODO: Alloc file context.

                return error.FileNotFound;
            },

            .native => |native| {
                if (native.len == 0) return error.FileNotFound;

                var path_buffer = [_]u8{0} ** 4096;
                const seperator_length = @boolToInt(native[native.len - 1] != Path.seperator);

                if ((native.len + seperator_length + path.length) >= path_buffer.len)
                    return error.FileNotFound;

                const io = core.io;

                io.copy(u8, &path_buffer, native);

                if (seperator_length != 0) path_buffer[native.len] = Path.seperator;

                io.copy(u8, path_buffer[native.len .. path_buffer.len],
                    path.buffer[0 .. path.length]);

                ext.SDL_ClearError();

                return @ptrCast(*ReadableFile, ext.SDL_RWFromFile(&path_buffer, "rb")
                    orelse return error.FileNotFound);
            },
        }
    }
};

///
///
///
pub const Graphics = opaque {
    ///
    ///
    ///
    pub const Event = struct {
        keys_up: Keys = std.mem.zeroes(Keys),
        keys_down: Keys = std.mem.zeroes(Keys),
        keys_held: Keys = std.mem.zeroes(Keys),

        const Keys = [256]bool;
    };

    ///
    ///
    ///
    const Implementation = struct {
        event: Event,
    };

    ///
    ///
    ///
    pub fn poll(graphics: *Graphics) ?*const Event {
        _ = graphics;

        return null;
    }

    ///
    ///
    ///
    pub fn present(graphics: *Graphics) void {
        // TODO: Implement;
        _ = graphics;
    }
};

///
/// [Logger.info] logs information that isn't necessarily an error but indicates something useful to
/// be logged.
///
/// [Logger.debug] logs information only when the engine is in debug mode.
///
/// [Logger.warning] logs information to indicate a non-critical error has occured.
///
pub const Logger = enum(u32) {
    info = ext.SDL_LOG_PRIORITY_INFO,
    debug = ext.SDL_LOG_PRIORITY_DEBUG,
    warning = ext.SDL_LOG_PRIORITY_WARN,
};

///
/// Linked list of asynchronous messages chained together to be processed by the work processor.
///
pub const Message = struct {
    next: ?*Message = null,

    kind: union(enum) {
        quit,

        log: struct {
            logger: Logger,
            info: []const u8,
        },

        task: struct {
            data: *anyopaque,
            action: fn (*anyopaque) void,
            frame: anyframe,
        },
    },
};

///
/// Path to a file on a [FileSystem].
///
pub const Path = extern struct {
    buffer: [255]u8,
    length: u8,

    ///
    /// [Error.TooLong] occurs when creating a path that is greater than the maximum path size **in
    /// bytes**.
    ///
    pub const FromError = error {
        TooLong,
    };

    ///
    /// An empty [Path] with a length of `0`.
    ///
    pub const empty = Path{
        .buffer = [_]u8{0} ** 255,
        .length = 0,
    };

    ///
    ///
    ///
    pub fn compare(this: Path, that: Path) isize {
        return core.io.compareBytes(this.buffer[0 ..this.length], that.buffer[0 .. that.length]);
    }

    ///
    /// Returns `true` if `this` is equal to `that`, otherwise `false`.
    ///
    pub fn equals(this: Path, that: Path) bool {
        return core.io.equals(u8, this.buffer[0 ..this.length], that.buffer[0 .. that.length]);
    }

    ///
    /// Attempts to create a [Path] with the path components in `sequences` as a fully qualified
    /// path from root.
    ///
    /// A [Path] value is returned containing the fully qualified path from the file-system root or
    /// a [FromError] if it could not be created.
    ///
    pub fn from(sequences: []const []const u8) FromError!Path {
        var path = empty;

        if (sequences.len != 0) {
            const last_sequence_index = sequences.len - 1;

            for (sequences) |sequence, index| if (sequence.len != 0) {
                var components = core.io.Spliterator(u8){
                    .source = sequence,
                    .delimiter = "/",
                };

                while (components.next()) |component| if (component.len != 0) {
                    for (component) |byte| {
                        if (path.length == max) return error.TooLong;

                        path.buffer[path.length] = byte;
                        path.length += 1;
                    }

                    if (components.hasNext()) {
                        if (path.length == max) return error.TooLong;

                        path.buffer[path.length] = '/';
                        path.length += 1;
                    }
                };

                if (index < last_sequence_index) {
                    if (path.length == max) return error.TooLong;

                    path.buffer[path.length] = '/';
                    path.length += 1;
                }
            };
        }

        return path;
    }

    ///
    /// Returns the hash of the text in `path`.
    ///
    pub fn hash(path: Path) usize {
        return core.io.hashBytes(path.buffer[0 .. path.length]);
    }

    ///
    /// Maximum number of **bytes** in a [Path].
    ///
    pub const max = 255;

    ///
    /// Textual separator between components of a [Path].
    ///
    pub const seperator = '/';
};

///
/// [RunError.InitFailure] occurs when the runtime fails to initialize.
///
pub const RunError = error {
    InitFailure,
};

///
/// Returns a [core.io.Allocator] bound to the underlying system allocator.
///
pub fn allocator() core.io.Allocator {
    // TODO: Add leak detection.
    return .{
        .bound = .{
            .alloc = ext.SDL_alloc,
            .dealloc = ext.SDL_free,
        },
    };
}

///
/// Runs a graphical application referenced by `run` with `error` as its error set.
///
/// Should an error from `run` occur, an `Error` is returned, otherwise a [RunError] is returned if
/// the underlying runtime fails and is logged.
///
pub fn display(comptime Error: anytype,
    comptime run: fn (*App, *Graphics) callconv(.Async) Error!void) (RunError || Error)!void {

    var cwd = FileSystem{.native = "./"};
    const user_prefix = ext.SDL_GetPrefPath("ona", "ona") orelse return error.InitFailure;

    defer ext.SDL_free(user_prefix);

    var gpa = std.heap.GeneralPurposeAllocator(.{}){};

    defer if (gpa.deinit())
        ext.SDL_LogWarn(ext.SDL_LOG_CATEGORY_APPLICATION, "Runtime allocator leaked memory");

    var app = App{
        .user = .{.native = std.mem.sliceTo(user_prefix, 0)},

        .message_semaphore = ext.SDL_CreateSemaphore(0) orelse {
            ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION,
                "Failed to create message semaphore");

            return error.InitFailure;
        },

        .message_mutex = ext.SDL_CreateMutex() orelse {
            ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION,
                "Failed to create message mutex");

            return error.InitFailure;
        },

        .data = .{
            .archive_file = cwd.openRead(try Path.from(&.{"./data.oar"})) catch {
                ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to load ./data.oar");

                return error.InitFailure;
            },
        },
    };

    defer {
        app.data.archive_file.close();
        ext.SDL_DestroySemaphore(app.message_semaphore);
        ext.SDL_DestroyMutex(app.message_mutex);
    }

    const message_thread = ext.SDL_CreateThread(processMessages, "Message Processor", &app) orelse {
        ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to create message processor");

        return error.InitFailure;
    };

    defer {
        var message = Message{.kind = .quit};

        app.enqueue(&message);

        {
            var status: c_int = 0;

            // SDL2 defines waiting on a null thread reference as a no-op. See
            // https://wiki.libsdl.org/SDL_WaitThread for more information
            ext.SDL_WaitThread(message_thread, &status);

            if (status != 0) {
                // TODO: Error check this.
            }
        }
    }

    if (ext.SDL_Init(ext.SDL_INIT_EVERYTHING) != 0) {
        ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to initialize runtime");

        return error.InitFailure;
    }

    defer ext.SDL_Quit();

    const window = create_window: {
        const pos = ext.SDL_WINDOWPOS_UNDEFINED;
        var flags = @as(u32, 0);

        break: create_window ext.SDL_CreateWindow("Ona", pos, pos, 640, 480, flags) orelse {
            ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to create window");

            return error.InitFailure;
        };
    };

    defer ext.SDL_DestroyWindow(window);

    const renderer = create_renderer: {
        var flags = @as(u32, 0);

        break: create_renderer ext.SDL_CreateRenderer(window, -1, flags) orelse {
            ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to create renderer");

            return error.InitFailure;
        };
    };

    defer ext.SDL_DestroyRenderer(renderer);

    var graphics = Graphics.Implementation{
        .event = .{

        },
    };

    return run(@ptrCast(*App, &app), @ptrCast(*Graphics, &graphics));
}

///
/// [FileSystemMessage] processing function used by a dedicated worker thread, where `data`
/// is a type-erased reference to a [EventLoop].
///
/// The processor returns `0` if it exited normally or any other value if an erroneous exit
/// occured.
///
pub fn processMessages(userdata: ?*anyopaque) callconv(.C) c_int {
    const app = @ptrCast(*App, @alignCast(@alignOf(App), userdata orelse unreachable));

    while (true) {
        // TODO: Error check these.
        _ = ext.SDL_SemWait(app.message_semaphore);
        _ = ext.SDL_LockMutex(app.message_mutex);

        defer _ = ext.SDL_UnlockMutex(app.message_mutex);

        while (app.message_chain) |message| {
            switch (message.kind) {
                .quit => return 0,

                .log => |log| ext.SDL_LogMessage(ext.SDL_LOG_CATEGORY_APPLICATION,
                    @enumToInt(log.logger), "%.*s", log.info.len, log.info.ptr),

                .task => |task| {
                    task.action(task.data);

                    resume task.frame;
                },
            }

            app.message_chain = message.next;
        }
    }
}
